# AUTOGENERATED BY NBDEV! DO NOT EDIT!

__all__ = ["index", "modules", "custom_doc_links", "git_url"]

index = {"func": "batch_process_segmentation.ipynb",
         "process_experiment_command": "batch_process_segmentation.ipynb",
         "process_experiment": "batch_process_segmentation.ipynb",
         "nlm_denoise": "denoise_images.ipynb",
         "doLap": "focus_stacking.ipynb",
         "focus_stack": "focus_stacking.ipynb",
         "require_jvm": "load_bioformats_images.ipynb",
         "get_image_reader": "load_bioformats_images.ipynb",
         "kill_jvm": "load_bioformats_images.ipynb",
         "load_slice": "load_tiff_images.ipynb",
         "load_meta": "load_bioformats_images.ipynb",
         "load_image": "load_tiff_images.ipynb",
         "load_hough_segmentation": "process_segmentation.ipynb",
         "compute_overlap": "process_segmentation.ipynb",
         "make_overlap_graph": "process_segmentation.ipynb",
         "plot_circles_on_img": "process_segmentation.ipynb",
         "lt_mean_filter": "process_segmentation.ipynb",
         "create_circular_mask": "process_segmentation.ipynb",
         "compute_patch_coordinates": "process_segmentation.ipynb",
         "extract_patch_from_image": "process_segmentation.ipynb",
         "extract_patches_from_image": "process_segmentation.ipynb",
         "find_hough_circle_per_node": "process_segmentation.ipynb",
         "hough_filter": "process_segmentation.ipynb",
         "erronous_segmentations_filter": "process_segmentation.ipynb"}

modules = ["analyse_intensity.py",
           "batch_process_segmentation.py",
           "denoise.py",
           "focus_stacking.py",
           "load_bioformats_images.py",
           "load_tiff_images.py",
           "process_segmentation.py"]

doc_url = "https://mlippeve.github.io/c7m/"

git_url = "https://github.com/mlippeve/c7m/tree/master/"

def custom_doc_links(name): return None



# AUTOGENERATED! DO NOT EDIT! File to edit: notebooks/denoise_images.ipynb (unless otherwise specified).

__all__ = ['nlm_denoise']

# Cell
# export
import pandas
import dotenv
import os
import matplotlib.pyplot as plt
import numpy

dotenv.load_dotenv()

# Cell

from c7m import load_tiff_images
import networkx
from tqdm.notebook import tqdm, trange
import click
import multiprocessing
import logging
import cv2 as cv

# Cell
def nlm_denoise(im):
    tmp = cv.convertScaleAbs(im, alpha=(255.0/im.max()))
    return cv.fastNlMeansDenoising(tmp, 3)


# AUTOGENERATED! DO NOT EDIT! File to edit: notebooks/focus_stacking.ipynb (unless otherwise specified).

__all__ = ['focus_stack']

# Cell
# export
import pandas
import dotenv
import os
import matplotlib.pyplot as plt
from importlib import reload
import numpy
import warnings

warnings.filterwarnings('error')

dotenv.load_dotenv()

# Cell
import cv2
from c7m import load_tiff_images
from skimage import img_as_float
from tqdm.notebook import tqdm
from PIL import Image

# Internal Cell

def doLap(image):
    kernel_size = 5         # Size of the laplacian window
    blur_size = 5           # How big of a kernal to use for the gaussian blur
                            # Generally, keeping these two values the same or very close works well
                            # Also, odd numbers, please...

    blurred = cv2.GaussianBlur(image, (blur_size,blur_size), 0)
    return cv2.Laplacian(blurred, cv2.CV_64F, ksize=kernel_size)

# Cell
def focus_stack(images):
    laps = []
    for i in range(len(images)):
        laps.append(doLap(images[i]))

    laps = numpy.asarray(laps)

    output = numpy.zeros(shape=images[0].shape, dtype=images[0].dtype)

    abs_laps = numpy.absolute(laps)
    maxima = abs_laps.max(axis=0)
    bool_mask = abs_laps == maxima
    mask = bool_mask.astype(numpy.uint8)
    for i in range(0,len(images)):
        output = cv2.bitwise_not(images[i],output, mask=mask[i])
    return cv2.bitwise_not(output)


# AUTOGENERATED! DO NOT EDIT! File to edit: notebooks/analyse_intensity.ipynb (unless otherwise specified).

__all__ = []

# Cell
# export
import pandas
import dotenv
import os
import matplotlib.pyplot as plt
import numpy
import scipy.stats, scipy.ndimage
from sklearn import mixture, ensemble, neighbors, exceptions
import warnings

dotenv.load_dotenv()

# Cell
from c7m import load_tiff_images, process_segmentation, denoise, focus_stacking
from skimage import img_as_float32
from tqdm.notebook import tqdm, trange
import seaborn
import cv2


# AUTOGENERATED! DO NOT EDIT! File to edit: notebooks/batch_process_segmentation.ipynb (unless otherwise specified).

__all__ = ['func', 'process_experiment_command', 'process_experiment']

# Cell
# export
import pandas
import dotenv
import os
import matplotlib.pyplot as plt
import numpy

dotenv.load_dotenv()

# Cell

from c7m import process_segmentation, load_tiff_images, focus_stacking
import networkx
from tqdm import tqdm, trange
import click
import multiprocessing
import logging
import cv2
from skimage import img_as_float32


# Cell

def func(i, series_circles, img):
    img = focus_stacking.focus_stack(img)
    img = cv2.bilateralFilter(img, 5, 10., 10.)

    g = process_segmentation.make_overlap_graph(series_circles)
    g_filtered = networkx.subgraph_view(g, filter_node=process_segmentation.lt_mean_filter(g))
    g_filtered = process_segmentation.compute_patch_coordinates(g_filtered)
    g_filtered = networkx.subgraph_view(g_filtered, filter_node=process_segmentation.erronous_segmentations_filter(g_filtered))
    g_filtered = process_segmentation.find_hough_circle_per_node(g_filtered, img, threshold=0.7)
    g_filtered = networkx.subgraph_view(g_filtered, filter_node=process_segmentation.hough_filter(g_filtered))

    data = numpy.empty((len(g_filtered),7), dtype=numpy.float64)
    for j,n in enumerate(g_filtered.nodes()):
        data[j, :4] = list(g_filtered.nodes[n]["patch_coordinates"])
        data[j, 4:] = list(g_filtered.nodes[n]["circle"])

    df = pandas.DataFrame.from_records(data, columns=["0", "1", "2", "3", "cx", "cy", "radius"])
    df["series"] = i

    return df


@click.command()
@click.option("--czi-path", type=click.Path(exists=True, dir_okay=False))
@click.option("--segmentation-csv", type=click.Path(exists=True, dir_okay=False))
@click.option("--output", type=click.Path(dir_okay=False), default=None, help="Path to pickle file to save segmentations.")
@click.option("--processes", type=int, default=6)
def process_experiment_command(*args, **kwargs):
    process_experiment(*args, **kwargs)

def process_experiment(path, segmentation_csv, output=None, processes=1, series_start=None, series_end=None):
    circles, series = process_segmentation.load_hough_segmentation(segmentation_csv)

    if series_end is None and series_start is None:
        r = series
    else:
        r = numpy.arange(series_start, series_end)

    with multiprocessing.Pool(processes) as pool:
        # loop over all series and submit to pool
        promises = []
        for i in r:
            if i not in series:
                print(f"series {i} is not present in the segmentation file")
                continue
            series_circles = circles[circles["series"] == i].reset_index(drop=True)
            img = img_as_float32(load_tiff_images.load_slice(c=1, series=i, z="all", path=path))
            promises.append(pool.apply_async(func, (i, series_circles, img)))

        # get results from the queue
        df_list = []
        for p in tqdm(promises):
            df_list.append(p.get())

    df = pandas.concat(df_list).reset_index(drop=True)
    if output is not None:
        df.to_pickle(output)

    return df


# AUTOGENERATED! DO NOT EDIT! File to edit: notebooks/load_bioformats_images.ipynb (unless otherwise specified).

__all__ = ['get_image_reader', 'kill_jvm', 'load_slice', 'load_meta']

# Cell
# export
import pandas
import dotenv
import os
import matplotlib.pyplot as plt
dotenv.load_dotenv()
import numpy

# Cell
import javabridge
import javabridge._javabridge as _jb
import bioformats

# Internal Cell
def require_jvm(func):
    def start_jvm(*args, **kwargs):
        if not _jb.get_vm().is_active():
            javabridge.start_vm(class_path=bioformats.JARS)
        return func(*args, **kwargs)
    return start_jvm

# Cell
@require_jvm
def get_image_reader(path, key=42):
    return bioformats.get_image_reader(key, path)

def kill_jvm():
    javabridge.kill_vm()

# Cell
@require_jvm
def load_slice(c=None, z=0, t=0, series=None, reader=None, path=None):
    if reader:
        return reader.read(c, z, t, series)
    elif path:
        with get_image_reader(path) as reader:
            if type(z) is str and z=="all":
                images = []
                for i in range(3):
                    images.append(reader.read(c, i, t, series))
                return images
            return reader.read(c, z, t, series)
    else:
        return None

@require_jvm
def load_meta(path):
    return bioformats.get_omexml_metadata(path)


# AUTOGENERATED! DO NOT EDIT! File to edit: notebooks/load_tiff_images.ipynb (unless otherwise specified).

__all__ = ['load_image', 'load_slice']

# Cell
# export
import pandas
import dotenv
import os
import matplotlib.pyplot as plt
dotenv.load_dotenv()
from PIL import Image
import numpy

# Cell

def load_image(path):
    """
    path - Path to the multipage-tiff file
    Returns all slices
    """
    img = Image.open(path)
    images = []
    for i in range(img.n_frames):
        img.seek(i)
        images.append(numpy.array(img))
    return numpy.array(images)

# Cell

def load_slice(c=None, z=0, t=0, series=None, path=None):
    """
    Return one slice from a multi-page tiff
    """

    n_slices = 3
    n_channels = 7

    if type(z) is not str:
        assert z in range(n_slices), f"There are only {n_slices} Z slices."
    assert c in range(n_channels), f"There are only {n_channels} channels."

    if series is None:
        img_path = path
    else:
        assert os.path.isdir(path), "Path is not found."
        img_path = os.path.join(path, f"series_{series}.ome.tiff")

    assert os.path.isfile(img_path), "Path is not found."

    with Image.open(img_path) as img:
        if (type(z) is str) and (z == "all"):
            images = []
            for z in range(n_slices):
                s = c+z*n_channels
                img.seek(s)
                images.append(numpy.array(img))
            return numpy.array(images, dtype=numpy.uint16)
        else:
            s = c + z*n_channels
            img.seek(s)
            return numpy.array(img, dtype=numpy.uint16)


# AUTOGENERATED! DO NOT EDIT! File to edit: notebooks/process_segmentation.ipynb (unless otherwise specified).

__all__ = ['load_hough_segmentation', 'make_overlap_graph', 'plot_circles_on_img', 'compute_patch_coordinates',
           'extract_patch_from_image', 'extract_patches_from_image', 'find_hough_circle_per_node', 'hough_filter',
           'erronous_segmentations_filter']

# Cell
# export
import pandas
import dotenv
import os
import matplotlib.pyplot as plt
from importlib import reload
import numpy
import warnings

warnings.filterwarnings('error')

dotenv.load_dotenv()

# Cell
from matplotlib import cm
import math
import itertools
import scipy.spatial
import networkx
from skimage.feature import canny
from skimage.transform import hough_circle, hough_circle_peaks
from skimage.draw import circle_perimeter
from skimage.filters import sobel
from skimage import img_as_float32
import re
import cv2

from c7m import load_tiff_images, denoise, focus_stacking

# Cell
def load_hough_segmentation(csv_file, micron_dims=(189.45, 189.45)):
    circles = pandas.read_csv(csv_file, index_col=False)
    circles = circles.drop(columns=["ID"])
    circles["X frac"] = circles["X (microns)"]/micron_dims[0]
    circles["Y frac"] = circles["Y (microns)"]/micron_dims[1]
    circles["Radius frac"] = circles["Radius (microns)"]/math.sqrt(micron_dims[0]*micron_dims[1])

    # extract series id from image title
    exp = r"^[^#]+#([0-9]+)-.+$"
    circles["series"] = circles["Image title"].apply(lambda a: int(re.match(exp, a).group(1))-1)
    return circles, circles["series"].unique()

# Internal Cell
def compute_overlap(d, r1, r2):
    # https://scipython.com/book/chapter-8-scipy/problems/p84/overlapping-circles/

    area = 0

    # only compute overlap if sum of radii is larger then distance between circle centers
    if r1 + r2 > d + 1e-5:
        r, R = (r1, r2) if r2 >= r1 else (r2, r1)

        try:
            alpha = numpy.arccos((r**2 + d**2 - R**2)/(2*r*d))
        except RuntimeWarning:
            print(r, R, d)
        beta = numpy.arccos((R**2 + d**2 - r**2)/(2*R*d))
        area = beta*R**2 + alpha*r**2 - (1/2)*r**2*numpy.sin(2*alpha) - (1/2)*R**2*numpy.sin(2*beta)

    return area

# Cell
def make_overlap_graph(df):
    g = networkx.Graph()

    # compute distance between all circle centers
    d = scipy.spatial.distance.squareform(scipy.spatial.distance.pdist(df[["X frac", "Y frac"]]))

    # select only those circle pairs whose center to center distance is larger then twice the smallest radius
    radius = df["Radius frac"].max()
    combinations = [c for c in itertools.combinations(df.index, 2) if d[c[0], c[1]] < radius*2]

    for idx1, idx2 in combinations:
        overlap = compute_overlap(d[idx1, idx2], df.iloc[idx1]["Radius frac"], df.iloc[idx2]["Radius frac"])

        g.add_node(idx1, **df.iloc[idx1].to_dict())
        g.add_node(idx2, **df.iloc[idx2].to_dict())

        if overlap > 0:
            g.add_edge(idx1, idx2, weight=overlap)

    return g

# Cell
def plot_circles_on_img(graph, img, highlight=[]):
    fig, ax = plt.subplots(figsize=(10, 10), dpi=100)

    ax.imshow(img, origin="lower", extent=(0, 1, 0, 1))
    ax.set_xlim(0, 1)
    ax.set_ylim(0, 1)

    for idx in graph.nodes:
        color = "orange"
        if idx in highlight:
            color="red"

        circle = graph.nodes[idx]

        c = plt.Circle((circle["X frac"], circle["Y frac"]), circle["Radius frac"], color=color, alpha=0.2)
        ax.add_artist(c)
        ax.annotate(str(idx), (circle["X frac"], circle["Y frac"]), ha="center", va="center")

    return ax

# Internal Cell

def lt_mean_filter(g):
    def func(n):
        if len(list(g.neighbors(n))) == 0:
            return True

        # keep this node if the mean score of its neighbors is lower then its score
        agg_neighbor_score = numpy.mean([g.nodes[k]["Score"] for k in g.neighbors(n)])
        if agg_neighbor_score < g.nodes[n]["Score"]:
            return True

        # check if total overlap with neighbors is less then 50% of its area
        tot_overlap = sum([g.get_edge_data(n, k)["weight"] for k in g.neighbors(n)])
        if tot_overlap < 0.5*numpy.pi*g.nodes[n]["Radius frac"]**2:
            return True

        return False
    return func

# Internal Cell

def create_circular_mask(h, w, center=None, radius=None):
    # https://stackoverflow.com/a/44874588/2977466

    if center is None: # use the middle of the image
        center = (int(w/2), int(h/2))
    if radius is None: # use the smallest distance between the center and image walls
        radius = min(center[0], center[1], w-center[0], h-center[1])

    Y, X = numpy.ogrid[:h, :w]
    dist_from_center = numpy.sqrt((X - center[0])**2 + (Y-center[1])**2)

    mask = dist_from_center <= radius
    return mask

# Cell

def compute_patch_coordinates(graph):
    for index in graph.nodes:
        circle = graph.nodes[index]

        x = circle["X frac"]
        y = circle["Y frac"]
        radius = circle["Radius frac"]

        circle["patch_coordinates"] = (max(0, y-radius), y+radius, max(0, x-radius), x+radius)
    return graph

def extract_patch_from_image(coords, img, masked=None, fill=0):
    size = img.shape[0]
    patch = img[
        int(coords[0]*size):int(coords[1]*size),
        int(coords[2]*size):int(coords[3]*size)
    ]
    if masked is not None:
        mask = create_circular_mask(patch.shape[0], patch.shape[1], (masked[0], masked[1]), masked[2])
        return numpy.where(mask, patch, fill)
    return patch

def extract_patches_from_image(graph, img, masked=False):
    patches = []

    for index in graph.nodes:
        coords = graph.nodes[index]["patch_coordinates"]

        m = None
        if masked:
            m = graph.nodes[index]["circle"]

        patch = extract_patch_from_image(coords, img, m)
        patches.append(patch)

    return patches

# Cell
def find_hough_circle_per_node(graph, img, threshold=0.7):
    for index in graph.nodes:
        coords = graph.nodes[index]["patch_coordinates"]
        size = img.shape[0]
        patch = img[
            int(coords[0]*size):int(coords[1]*size),
            int(coords[2]*size):int(coords[3]*size)
        ]

        edges = sobel(patch)
        bin_edges = numpy.where(edges > numpy.mean(edges), 1, 0)
        try_radii = numpy.arange(min(patch.shape)*0.3, min(patch.shape)*0.5)
        result = hough_circle(bin_edges, try_radii)
        accums, cx, cy, radii = hough_circle_peaks(result, try_radii, total_num_peaks=1, threshold=threshold)

        if len(accums) > 0:
            max_idx = numpy.argmax(accums)
            graph.nodes[index]["circle"] = cx[max_idx], cy[max_idx], radii[max_idx]
        else:
            graph.nodes[index]["circle"] = None

    return graph

# Cell
def hough_filter(graph):
    def func(index):
        if graph.nodes[index]["circle"] is not None:
            return True

    return func

# Cell
def erronous_segmentations_filter(graph):
    def func(index):
        coords = graph.nodes[index]["patch_coordinates"]

        # check if width and height aren't equal
        if numpy.abs(coords[0]-coords[1]) < 1e-3:
            return False
        if numpy.abs(coords[2]-coords[3]) < 1e-3:
            return False

        return True
    return func